﻿// Disclaimer:
// This software is provided AS IS for educational purposes only,
// with no warranties, and confers no rights. Error checking is 
// elided for clarity. Carry out sufficient tests prior to using
// it in a productive environment.
// CASIO Europe GmbH accepts no responsibility for loss or damage
// arising from its use, including damage from virus.
//
// Please, note: This configuration information may be changed
// without any further notice.
// Casio Europe GmbH does not guarantee future compatibility. By 
// using this software you agree to these terms and conditions.
//
// Version 1.04 - April 05, 2012
// Supported Devices: DT-X7, DT-X8, DT-X11, DT-X30, IT-300, IT-600, 
// IT-800, IT-3100, and IT-9000.
//
// Remarks: You need latest versions of CASIO runtime libraries on
// your mobile device. If they are not already installed, you can 
// find them as part of the SDK.
//////////////////////////////////////////////////////////////////////

using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Calib;
using System.IO;
using System.Diagnostics;
using System.Threading;
using System.Runtime.InteropServices;

namespace BTSerial
{
    public enum TargetPlatform
    {
        None,
        WindowsCE,
        WindowsMobile
    }

    public partial class MainForm : Form
    {
        #region Fields

        private Action<string> onDataReceived;
        private PowerBroadcast powerBroadcast;
        private bool btInitialized;
        private ILogFile logFile;
        private string passKey;
        private TargetPlatform target;
        private int scaling;

        #endregion

        #region Default Settings

        private const int InquiryTimeOut = 10000;
        private const int MaxReceivedChars = 100;
        private const string PassKeyDefault = "0000"; // Fits most cases
        private const string EndOfLine = "\r\n";
        private const int SysErrSuccess = -1;

        private readonly Color wmConnectEnabled = Color.Red;
        private readonly Color wmConnectDisabled = Color.LightSalmon;
        private readonly Color wmDisconnectEnabled = Color.Blue;
        private readonly Color wmDisconnectDisabled = Color.LightBlue;

        #endregion

        #region Properties

        public TargetPlatform Target
        {
            get { return this.target; }
        }

        #endregion

        #region Init / Deinit

        public MainForm()
        {
            InitializeComponent();

            // By default ASCIIEncoding is used.
            // Change here if necessary, e.g.
            //UnicodeEncoding utf16 = new UnicodeEncoding();
            //this.serialPort.Encoding = utf16;

            WindowState = FormWindowState.Maximized;
            this.btInitialized = false;
            this.powerBroadcast = new PowerBroadcast(this);
            this.logFile = new BTSerialLogFile(Name);
            this.onDataReceived = s => this.tbxReceived.Text += s;
            this.passKey = PassKeyDefault;
            SysGetModelName();
            SysDisablePowerOff();
            BTInitialize();
            BTInitializeInfos();
            CheckStatus();
        }

        private void MainForm_Closing(object sender, CancelEventArgs e)
        {
            Cursor.Current = Cursors.WaitCursor;
            SysEnablePowerOff();
            SerialPortClose();
            BTDeinitialize();
            Cursor.Current = Cursors.Default;
        }

        #endregion

        #region Overrides

        protected override void ScaleControl(SizeF factor, BoundsSpecified specified)
        {
            if (factor.ToSize().Width == 2)
            {
                this.scaling = 2;
                this.lsvDevices.SmallImageList = this.imgSmallIconsVga;
            }
            else
            {
                this.scaling = 1;
                this.lsvDevices.SmallImageList = this.imgSmallIconsQvga;
            }

            base.ScaleControl(factor, specified);
        }

        #endregion

        #region Button Handlers

        private void btnInquiry_Click(object sender, EventArgs e)
        {
            if (this.serialPort.IsOpen == true)
            {
                this.statusBar.Text = Resources.StatusSerialPortIsOpened;
                return;
            }
            if (MessageForm.ShowDialogEx(Resources.MsgBTNewInquiry, this) == DialogResult.OK)
            {
                Refresh();
                Cursor.Current = Cursors.WaitCursor;
                if (BTInquiry() == true)
                {
                    this.statusBar.Text = Resources.StatusBTInquiryOK;
                }
                else
                {
                    this.statusBar.Text = Resources.StatusBTInquiryNG;
                }
                Cursor.Current = Cursors.Default;
            }
        }

        private void btnConnect_Click(object sender, EventArgs e)
        {
            if (this.lsvDevices.FocusedItem == null)
            {
                this.statusBar.Text = Resources.StatusNoDeviceSelected;
                return;
            }
            if (this.serialPort.IsOpen == true)
            {
                this.statusBar.Text = Resources.StatusSerialPortIsOpened;
                return;
            }

            Cursor.Current = Cursors.WaitCursor;
            if (BTConnect() == true)
            {
                EnableControls(true);
                this.statusBar.Text = Resources.StatusBTConnectionOK;
            }
            else
            {
                this.statusBar.Text = Resources.StatusBTConnectionNG;
            }
            Cursor.Current = Cursors.Default;
        }

        private void btnDisconnect_Click(object sender, EventArgs e)
        {
            if (this.serialPort.IsOpen == false)
            {
                this.statusBar.Text = Resources.StatusSerialPortIsClosed;
                return;
            }

            Cursor.Current = Cursors.WaitCursor;
            if (SerialPortClose() == true)
            {
                EnableControls(false);
                this.statusBar.Text = Resources.StatusBTDisconnectionOK;
            }
            else
            {
                this.statusBar.Text = Resources.StatusBTDisconnectionNG;
            }
            Cursor.Current = Cursors.Default;
        }

        private void btnSend_Click(object sender, EventArgs e)
        {
            if (this.serialPort.IsOpen == false)
            {
                this.statusBar.Text = Resources.StatusSerialPortIsClosed;
                return;
            }

            Cursor.Current = Cursors.WaitCursor;
            if (BTSend() == true)
            {
                this.statusBar.Text = Resources.StatusSerialPortSendOK;
            }
            else
            {
                PlayBeep();
                if (SerialPortClose() == true)
                {
                    EnableControls(false);
                    this.statusBar.Text = Resources.StatusSerialPortSendNG;
                }
                else
                {
                    this.statusBar.Text = Resources.StatusSerialPortSendNGError;
                }
            } 
            Cursor.Current = Cursors.Default;
        }

        private void btnClear_Click(object sender, EventArgs e)
        {
            this.tbxReceived.Text = "";
        }

        private void btnSetDefault_Click(object sender, EventArgs e)
        {
            if (this.lsvDevices.FocusedItem == null)
            {
                this.statusBar.Text = Resources.StatusNoDeviceSelected;
                return;
            }
            if (this.serialPort.IsOpen == true)
            {
                this.statusBar.Text = Resources.StatusSerialPortIsOpened;
                return;
            }

            Cursor.Current = Cursors.WaitCursor;
            if (BTSetDefault() == true)
            {
                this.statusBar.Text = Resources.StatusBTSetDefaultOK;
            }
            else
            {
                this.statusBar.Text = Resources.StatusBTSetDefaultNG;
            }
            Cursor.Current = Cursors.Default;
        }

        private void btnSetKey_Click(object sender, EventArgs e)
        {
            if (this.serialPort.IsOpen == true)
            {
                this.statusBar.Text = Resources.StatusSerialPortIsOpened;
                return;
            }

            if (PassKeyForm.ShowDialogEx(this.passKey, this) == DialogResult.OK)
            {
                this.passKey = PassKeyForm.PassKey;
                if (BTSetKey() == true)
                {
                    this.statusBar.Text = Resources.StatusBTSetKeyOK;
                }
                else
                {
                    this.statusBar.Text = Resources.StatusBTSetKeyNG;
                }
            }
        }

        private void btnTrust_Click(object sender, EventArgs e)
        {
            if (this.lsvDevices.FocusedItem == null)
            {
                this.statusBar.Text = Resources.StatusNoDeviceSelected;
                return;
            }
            if (this.serialPort.IsOpen == true)
            {
                this.statusBar.Text = Resources.StatusSerialPortIsOpened;
                return;
            }

            Cursor.Current = Cursors.WaitCursor;
            if (BTTrust() == true)
            {
                this.statusBar.Text = Resources.StatusBTTrustOK;
            }
            else
            {
                this.statusBar.Text = Resources.StatusBTTrustNG;
            }
            Cursor.Current = Cursors.Default;
        }

        private void btnOK_Click(object sender, EventArgs e)
        {
            Close();
        }

        #endregion

        #region Other Handlers

        private void MainForm_KeyUp(object sender, KeyEventArgs e)
        {
            if (this.target == TargetPlatform.WindowsMobile)
            {
                switch (e.KeyCode)
                {
                    case Keys.F1:
                        btnConnect_Click(this, new EventArgs());
                        break;
                    case Keys.F2:
                        btnDisconnect_Click(this, new EventArgs());
                        break;
                    case Keys.F4:
                        Close();
                        break;
                    default:
                        break;
                }
            }
            else if (e.KeyCode == Keys.Escape)
            {
                Close();
            }
        }

        private void tbxSend_GotFocus(object sender, EventArgs e)
        {
            if (this.serialPort.IsOpen == true)
            {
                this.inputPanel.Enabled = true;
            }
        }

        private void tbxSend_LostFocus(object sender, EventArgs e)
        {
            this.inputPanel.Enabled = false;
        }

        private void menuDetails_Click(object sender, EventArgs e)
        {
            ListViewItem lvi = this.lsvDevices.FocusedItem;

            if (lvi != null)
            {
                BluetoothLibNet.BTST_DEVICEINFO info = (BluetoothLibNet.BTST_DEVICEINFO)lvi.Tag;
                StringBuilder temp = new StringBuilder();
                temp.AppendFormat("MAC {0}, Profiles:", lvi.Text);
                for (int i = 0; i < info.ProfileNumber; i++)
                {
                    switch (info.ProfileUUID[i])
                    {
                        case BluetoothLibNet.Def.BTUUID_BASIC_PRINTING:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_BASIC_PRINTING);
                            break;
                        case BluetoothLibNet.Def.BTUUID_BNEP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_BNEP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_BROWSE_GROUP_DESCRIPTOR:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_BROWSE_GROUP_DESCRIPTOR);
                            break;
                        case BluetoothLibNet.Def.BTUUID_CHR_PRINT:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_CHR_PRINT);
                            break;
                        case BluetoothLibNet.Def.BTUUID_COMMON_ISDN_ACCESS:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_COMMON_ISDN_ACCESS);
                            break;
                        case BluetoothLibNet.Def.BTUUID_CORDLESS_TELEPHONY:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_CORDLESS_TELEPHONY);
                            break;
                        case BluetoothLibNet.Def.BTUUID_CTP:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_CTP);
                            break;
                        case BluetoothLibNet.Def.BTUUID_CTP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_CTP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_DIALUP_NETWORKING:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_DIALUP_NETWORKING);
                            break;
                        case BluetoothLibNet.Def.BTUUID_DIRECT_PRINTING:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_DIRECT_PRINTING);
                            break;
                        case BluetoothLibNet.Def.BTUUID_DIRECT_PRINTING_REFERENCE_OBJECTS:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_DIRECT_PRINTING_REFERENCE_OBJECTS);
                            break;
                        case BluetoothLibNet.Def.BTUUID_FAX:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_FAX);
                            break;
                        case BluetoothLibNet.Def.BTUUID_FTP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_FTP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_GENERIC_AUDIO:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_GENERIC_AUDIO);
                            break;
                        case BluetoothLibNet.Def.BTUUID_GENERIC_FILE_TRANSFER:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_GENERIC_FILE_TRANSFER);
                            break;
                        case BluetoothLibNet.Def.BTUUID_GENERIC_NETWORKING:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_GENERIC_NETWORKING);
                            break;
                        case BluetoothLibNet.Def.BTUUID_GENERIC_TELEPHONY:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_GENERIC_TELEPHONY);
                            break;
                        case BluetoothLibNet.Def.BTUUID_GN:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_GN);
                            break;
                        case BluetoothLibNet.Def.BTUUID_HANDSFREE:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_HANDSFREE);
                            break;
                        case BluetoothLibNet.Def.BTUUID_HANDSFREE_AUDIO_GATEWAY:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_HANDSFREE_AUDIO_GATEWAY);
                            break;
                        case BluetoothLibNet.Def.BTUUID_HARDCOPY_CABLE_REPLACEMENT:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_HARDCOPY_CABLE_REPLACEMENT);
                            break;
                        case BluetoothLibNet.Def.BTUUID_HCR_SCAN:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_HCR_SCAN);
                            break;
                        case BluetoothLibNet.Def.BTUUID_HEADSET:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_HEADSET);
                            break;
                        case BluetoothLibNet.Def.BTUUID_HEADSET_AUDIO_GATEWAY:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_HEADSET_AUDIO_GATEWAY);
                            break;
                        case BluetoothLibNet.Def.BTUUID_HIDP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_HIDP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_HTTP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_HTTP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_HUMAN_INTERFACE_DEVICE_SERVICE:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_HUMAN_INTERFACE_DEVICE_SERVICE);
                            break;
                        case BluetoothLibNet.Def.BTUUID_IMAGING:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_IMAGING);
                            break;
                        case BluetoothLibNet.Def.BTUUID_IMAGING_AUTOMATIC_ARCHIVE:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_IMAGING_AUTOMATIC_ARCHIVE);
                            break;
                        case BluetoothLibNet.Def.BTUUID_IMAGING_REFERENCE_OBJECTS:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_IMAGING_REFERENCE_OBJECTS);
                            break;
                        case BluetoothLibNet.Def.BTUUID_IMAGING_RESPONDER:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_IMAGING_RESPONDER);
                            break;
                        case BluetoothLibNet.Def.BTUUID_INTERCOM:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_INTERCOM);
                            break;
                        case BluetoothLibNet.Def.BTUUID_IP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_IP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_IR_MC_SYNC:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_IR_MC_SYNC);
                            break;
                        case BluetoothLibNet.Def.BTUUID_IR_MC_SYNC_COMMAND:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_IR_MC_SYNC_COMMAND);
                            break;
                        case BluetoothLibNet.Def.BTUUID_L2CAP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_L2CAP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_LAN_ACCESS_USING_PPP:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_LAN_ACCESS_USING_PPP);
                            break;
                        case BluetoothLibNet.Def.BTUUID_NAP:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_NAP);
                            break;
                        case BluetoothLibNet.Def.BTUUID_OBEX_FILE_TRANSFER:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_OBEX_FILE_TRANSFER);
                            break;
                        case BluetoothLibNet.Def.BTUUID_OBEX_OBJECT_PUSH:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_OBEX_OBJECT_PUSH);
                            break;
                        case BluetoothLibNet.Def.BTUUID_OBEX_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_OBEX_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_PANU:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_PANU);
                            break;
                        case BluetoothLibNet.Def.BTUUID_PNP_INFORMATION:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_PNP_INFORMATION);
                            break;
                        case BluetoothLibNet.Def.BTUUID_PRINTING_STATUS:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_PRINTING_STATUS);
                            break;
                        case BluetoothLibNet.Def.BTUUID_PUBLIC_BROWSE_ROOT:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_PUBLIC_BROWSE_ROOT);
                            break;
                        case BluetoothLibNet.Def.BTUUID_REFERENCE_PRINTING:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_REFERENCE_PRINTING);
                            break;
                        case BluetoothLibNet.Def.BTUUID_REFLECTED_UI:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_REFLECTED_UI);
                            break;
                        case BluetoothLibNet.Def.BTUUID_RFCOMM_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_RFCOMM_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_SDP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_SDP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_SERIAL_PORT:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_SERIAL_PORT);
                            break;
                        case BluetoothLibNet.Def.BTUUID_SERVICE_DISCOVERY_SERVER:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_SERVICE_DISCOVERY_SERVER);
                            break;
                        case BluetoothLibNet.Def.BTUUID_TCP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_TCP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_TCS_AT_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_TCS_AT_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_TCS_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_TCS_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_UDP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_UDP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_UPNP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_UPNP_PROTOCOL);
                            break;
                        case BluetoothLibNet.Def.BTUUID_VIDEO_CONFERENCING_GW:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_VIDEO_CONFERENCING_GW);
                            break;
                        case BluetoothLibNet.Def.BTUUID_WAP:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_WAP);
                            break;
                        case BluetoothLibNet.Def.BTUUID_WAP_CLIENT:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_WAP_CLIENT);
                            break;
                        case BluetoothLibNet.Def.BTUUID_WSP_PROTOCOL:
                            temp.AppendFormat("\n- {0}", Resources.BTUUID_WSP_PROTOCOL);
                            break;
                        default:
                            break;
                    }
                }
                MessageForm.ShowDialogEx(temp.ToString(), this);
            }
        }

        private void lsvDevices_ItemActivate(object sender, EventArgs e)
        {
            if (this.target == TargetPlatform.WindowsCE)
            {
                menuDetails_Click(this, new EventArgs());
            }
        }

        #endregion

        #region BT Init / Deinit

        private void BTInitialize()
        {
            if (this.btInitialized == true)
            {
                return;
            }

            try
            {
                int result = BluetoothLibNet.Api.BTInitialize();
                if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                {
                    this.logFile.Append("BTInitialize: " + result.ToString());
                }
                else
                {
                    this.btInitialized = true;
                }

                result = BluetoothLibNet.Api.BTSetPassKey(passKey);
                if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                {
                    this.logFile.Append("BTSetPassKey: " + result.ToString());
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        private void BTInitializeInfos()
        {
            if (this.btInitialized == false)
            {
                return;
            }

            try
            {
                int result;
                BluetoothLibNet.BTST_LOCALINFO li = new BluetoothLibNet.BTST_LOCALINFO();
                result = BluetoothLibNet.Api.BTGetLocalInfo(li);
                if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                {
                    this.lblMac.Text = Resources.MsgMacUnknown;
                    this.logFile.Append("BTGetLocalInfo: " + result.ToString());
                }
                else
                {
                    this.lblMac.Text = Resources.MsgMacKnown + 
                        new string(li.LocalAddress.TakeWhile(c => c != '\0').ToArray());
                    li.Authentication = false;
                    li.Encryption = false;
                    result = BluetoothLibNet.Api.BTSetLocalInfo(li);
                    if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                    {
                        this.logFile.Append("BTSetLocalInfo: " + result.ToString());
                    }

                    int number = 0;
                    result = BluetoothLibNet.Api.BTSearchDeviceInfo(IntPtr.Zero, ref number, IntPtr.Zero);
                    if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                    {
                        this.logFile.Append("BTSearchDeviceInfo: " + result.ToString());
                    }
                    else if (number > 0)
                    {
                        BluetoothLibNet.BTST_DEVICEINFO[] infos = CreateDeviceInfos(number + 1);
                        result = BluetoothLibNet.Api.BTSearchDeviceInfo(infos, ref number, IntPtr.Zero);
                        if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                        {
                            this.logFile.Append("BTSearchDeviceInfo: " + result.ToString());
                        }
                        else
                        {
                            for (int i = 0; i < number; i++)
                            {
                                if (infos[i].ProfileUUID.Any(uuid => uuid == BluetoothLibNet.Def.BTUUID_SERIAL_PORT) == true)
                                {
                                    AddInfo(infos[i]);
                                }
                            }
                            SetColumnWidth();
                        }
                    }
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        private void BTDeinitialize()
        {
            if (this.btInitialized == false)
            {
                return;
            }

            try
            {
                int result = BluetoothLibNet.Api.BTDeInitialize();
                if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                {
                    this.logFile.Append("BTDeInitialize: " + result.ToString());
                }
                else
                {
                    this.btInitialized = false;
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        #endregion

        #region BT Administration

        private bool BTSetDefault()
        {
            bool success = false;

            if (this.btInitialized == false || 
                this.serialPort.IsOpen == true ||
                this.lsvDevices.FocusedItem == null)
            {
                return false;
            }

            try
            {
                BluetoothLibNet.BTST_DEVICEINFO info =
                    (BluetoothLibNet.BTST_DEVICEINFO)this.lsvDevices.FocusedItem.Tag;
                if(info != null)
                {
                    int result = BluetoothLibNet.Api.BTSetDefaultDevice(
                        info,
                        BluetoothLibNet.Def.BTPORT_SERIAL);
                    if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                    {
                        this.logFile.Append("BTSetDefaultDevice: " + result.ToString());
                    }
                    else
                    {
                        success = true;
                    }
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }

            return success;
        }

        private bool BTSetKey()
        {
            bool success = false;

            if (this.btInitialized == false ||
                this.serialPort.IsOpen == true)
            {
                return false;
            }

            try
            {
                int result = BluetoothLibNet.Api.BTSetPassKey(this.passKey);
                if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                {
                    this.logFile.Append("BTSetPassKey: " + result.ToString());
                }
                else
                {
                    success = true;
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }

            return success;
        }

        private bool BTTrust()
        {
            bool success = false;

            if (this.btInitialized == false ||
                this.serialPort.IsOpen == true ||
                this.lsvDevices.FocusedItem == null)
            {
                return false;
            }

            try
            {
                BluetoothLibNet.BTST_DEVICEINFO info =
                    (BluetoothLibNet.BTST_DEVICEINFO)this.lsvDevices.FocusedItem.Tag;
                if(info != null)
                {
                    int result = BluetoothLibNet.Api.BTTrustDevice(
                        info,
                        BluetoothLibNet.Def.BTPORT_SERIAL);
                    if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                    {
                        this.logFile.Append("BTTrustDevice: " + result.ToString());
                    }
                    else
                    {
                        success = true;
                    }
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }

            return success;
        }

        private BluetoothLibNet.BTST_DEVICEINFO BTFindInfo(string mac)
        {
            BluetoothLibNet.BTST_DEVICEINFO info = null;

            try
            {
                int number = 1;
                BluetoothLibNet.BTST_DEVICEINFO[] infos = CreateDeviceInfos();
                int result = BluetoothLibNet.Api.BTSearchDeviceInfo(infos, ref number, mac);
                if (result != BluetoothLibNet.Def.BTERR_SUCCESS || number != 1)
                {
                    this.logFile.Append("BTSearchDeviceInfo: " + result.ToString());
                }
                else
                {
                    info = infos[0];
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
                
            return info;
        }

        #endregion

        #region BT Communications

        private bool BTInquiry()
        {
            bool success = false;
            bool error = false;

            if (this.btInitialized == false ||
                this.serialPort.IsOpen == true)
            {
                return false;
            }

            try
            {
                BTDeleteDeviceInfos();
                int number = 0;
                int result = BluetoothLibNet.Api.BTInquiry(IntPtr.Zero, ref number, InquiryTimeOut);
                if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                {
                    error = true;
                    this.logFile.Append("BTInquiry: " + result.ToString());
                }
                else
                {
                    if (number > 0)
                    {
                        BluetoothLibNet.BTST_DEVICEINFO[] infos = CreateDeviceInfos(number);
                        result = BluetoothLibNet.Api.BTGetDeviceInfo(infos, number, 0);
                        if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                        {
                            error = true;
                            this.logFile.Append("BTGetDeviceInfo: " + result.ToString());
                        }
                        else
                        {
                            foreach (BluetoothLibNet.BTST_DEVICEINFO info in infos)
                            {
                                result = BluetoothLibNet.Api.BTGetServiceInfo(info);
                                if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                                {
                                    Debug.WriteLine("BTGetServiceInfo: " + result.ToString());
                                }
                                else
                                {
                                    if (info.ProfileUUID.Any(uuid => uuid == BluetoothLibNet.Def.BTUUID_SERIAL_PORT) == true)
                                    {
                                        result = BluetoothLibNet.Api.BTRegisterDeviceInfo(info);
                                        if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                                        {
                                            error = true;
                                            this.logFile.Append("BTRegisterDeviceInfo: " + result.ToString());
                                        }
                                        else
                                        {
                                            AddInfo(info);
                                        }
                                    }
                                }
                            }
                            SetColumnWidth();
                        }
                    }
                }
                success = (error == false);
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }

            return success;
        }

        private void BTDeleteDeviceInfos()
        {
            try
            {
                foreach (ListViewItem lvi in this.lsvDevices.Items)
                {
                    BluetoothLibNet.BTST_DEVICEINFO info = BTFindInfo(lvi.Text);
                    if (info != null)
                    {
                        int result = BluetoothLibNet.Api.BTDeleteDeviceInfo(info);
                        if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                        {
                            this.logFile.Append("BTDeleteDeviceInfo: " + result.ToString());
                        }
                    }
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            finally
            {
                this.lsvDevices.Items.Clear();
            }
        }

        private bool BTConnect()
        {
            bool success = false;

            if (this.btInitialized == false || 
                this.serialPort.IsOpen == true ||
                this.lsvDevices.FocusedItem == null)
            {
                return false;
            }

            try
            {
                BluetoothLibNet.BTST_DEVICEINFO info =
                    (BluetoothLibNet.BTST_DEVICEINFO)this.lsvDevices.FocusedItem.Tag;
                if (info != null)
                {
                    int result = BluetoothLibNet.Api.BTSelectDevice(
                        info, 
                        BluetoothLibNet.Def.BTPORT_SERIAL);
                    if (result != BluetoothLibNet.Def.BTERR_SUCCESS)
                    {
                        this.logFile.Append("BTSelectDevice: " + result.ToString());
                    }
                    else
                    {
                        this.serialPort.Open();
                        if (this.serialPort.IsOpen != true)
                        {
                            this.logFile.Append("Serial port not opened");
                        }
                        else
                        {
                            success = true;
                        }
                    }
                }
            }
            catch (InvalidOperationException exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            catch (ArgumentOutOfRangeException exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            catch (ArgumentException exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            catch (IOException exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            catch (UnauthorizedAccessException exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }

            return success;
        }

        private bool SerialPortClose()
        {
            bool success = false;

            try
            {
                if (this.serialPort.IsOpen == true)
                {
                    this.serialPort.Close();
                }

                success = (this.serialPort.IsOpen == false);
            }
            catch (InvalidOperationException exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }

            return success;
        }

        private bool BTSend()
        {
            bool success = false;

            try
            {
                if (this.serialPort.IsOpen == true)
                {
                    this.serialPort.Write(
                        string.Format(
                        "{0}{1}",
                        this.tbxSend.Text,
                        (this.chkReturnNewLine.Checked == true) ? EndOfLine : ""));

                    success = true;
                }
            }
            catch (InvalidOperationException exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            catch (ArgumentNullException exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            catch (TimeoutException exc)
            {
                Debug.WriteLine(exc.ToString());
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }

            return success;
        }

        private void serialPort_DataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            try
            {
                if (e.EventType == System.IO.Ports.SerialData.Chars)
                {
                    string data = this.serialPort.ReadExisting();
                    if (data.Length > MaxReceivedChars)
                    {
                        data = data.Substring(data.Length - MaxReceivedChars);
                    }
                    BeginInvoke(onDataReceived, new object[] { data });
                }
            }
            catch(Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        private void serialPort_ErrorReceived(object sender, System.IO.Ports.SerialErrorReceivedEventArgs e)
        {
            this.logFile.Append("Serial Port: " + e.EventType.ToString());
        }

        private void tbxReceived_TextChanged(object sender, EventArgs e)
        {
            try
            {
                string temp = this.tbxReceived.Text;
                while (temp.Length > MaxReceivedChars)
                {
                    int ndx = temp.IndexOf(EndOfLine);
                    if (ndx != -1)
                    {
                        temp = temp.Remove(0, ndx + EndOfLine.Length);
                    }
                    else
                    {
                        temp = temp.Substring(MaxReceivedChars / 2);
                        break;
                    }
                }
                this.tbxReceived.Text = temp;

                if (this.tbxReceived.TextLength > 0)
                {
                    this.tbxReceived.Select(this.tbxReceived.TextLength, 0);
                    this.tbxReceived.ScrollToCaret();
                    this.btnClear.Enabled = true;
                }
                else
                {
                    this.btnClear.Enabled = false;
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        #endregion

        #region PowerBroadcast

        internal void OnSuspend()
        {
            Cursor.Current = Cursors.WaitCursor;
            if (SerialPortClose() == true)
            {
                BTDeinitialize();
                SysPowerOff();
            }
            Cursor.Current = Cursors.Default;
        }

        internal void OnResumeSuspend()
        {
            Cursor.Current = Cursors.WaitCursor;
            this.statusBar.Text = "";
            if (SerialPortClose() == true)  // fallback for special cases, e.g. VDET2
            {
                BTDeinitialize();
                Thread.Sleep(5000);
                SysDisablePowerOff();
                BTInitialize();
                CheckStatus();
            }
            Cursor.Current = Cursors.Default;
        }

        internal void OnBatteryLow()
        {
            PlayBeep();
            OnSuspend();
        }

        #endregion

        #region System Library

        private void SysGetModelName()
        {
            try
            {
                int model = 0;
                int version = 0;
                int platform = 0;
                int result = SystemLibNet.Api.SysGetModelName(
                    ref model,
                    ref version,
                    ref platform);
                if (result != SysErrSuccess)
                {
                    this.logFile.Append("SysGetModelName: " + result.ToString());
                }
                else
                {
                    switch (model)
                    {
                        case 4:     // IT-3100
                            this.target = TargetPlatform.WindowsCE;
                            this.serialPort.PortName = "COM5";
                            break;
                        case 7:     // IT-600
                        case 8:     // DT-X11
                        case 9:     // DT-X7
                        case 10:    // DT-X30CE
                        case 16:    // DT-X8
                        case 19:    // IT-9000CE
                            this.target = TargetPlatform.WindowsCE;
                            this.serialPort.PortName = "COM6";
                            break;
                        case 11:    // DT-X30WM
                        case 13:    // IT-800
                        case 15:    // IT-300
                        case 20:    // IT-9000WM
                            this.target = TargetPlatform.WindowsMobile;
                            this.serialPort.PortName = "COM6";
                            this.btnOK.BackColor = Color.Yellow;
                            break;
                        default:
                            this.target = TargetPlatform.None;
                            this.serialPort.PortName = "COM6";
                            break;
                    }
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        private void SysPowerOff()
        {
            try
            {
                int result = SystemLibNet.Api.SysPowerOff();
                if (result != SysErrSuccess)
                {
                    this.logFile.Append("SysPowerOff: " + result.ToString());
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        private void SysEnablePowerOff()
        {
            try
            {
                int result = SystemLibNet.Api.SysEnablePowerOff();
                if (result != SysErrSuccess)
                {
                    this.logFile.Append("SysEnablePowerOff: " + result.ToString());
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        private void SysDisablePowerOff()
        {
            try
            {
                int result = SystemLibNet.Api.SysDisablePowerOff();
                if (result != SysErrSuccess)
                {
                    this.logFile.Append("SysDisablePowerOff: " + result.ToString());
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        #endregion

        #region Helper Methods

        private void EnableControls(bool connected)
        {
            if (connected == true)
            {
                this.btnInquiry.Enabled = false;
                this.btnConnect.Enabled = false;
                this.btnDisconnect.Enabled = true;
                this.btnSetDefault.Enabled = false;
                this.btnSetKey.Enabled = false;
                this.btnTrust.Enabled = false;
                this.tbxSend.Enabled = true;
                this.chkReturnNewLine.Enabled = true;
                this.btnSend.Enabled = true;
                this.lsvDevices.Focus();
            }
            else
            {
                this.btnInquiry.Enabled = true;
                this.btnConnect.Enabled = true;
                this.btnDisconnect.Enabled = false;
                this.btnSetDefault.Enabled = true;
                this.btnSetKey.Enabled = true;
                this.btnTrust.Enabled = true;
                this.tbxSend.Enabled = false;
                this.chkReturnNewLine.Enabled = false;
                this.btnSend.Enabled = false;
                this.btnConnect.Focus();
            }

            if (this.target == TargetPlatform.WindowsMobile)
            {
                this.btnConnect.BackColor =
                    (this.btnConnect.Enabled == true) ?
                    this.wmConnectEnabled :
                    this.wmConnectDisabled;
                this.btnDisconnect.BackColor =
                    (this.btnDisconnect.Enabled == true) ?
                    this.wmDisconnectEnabled :
                    this.wmDisconnectDisabled;
            }
        }

        private void SetColumnWidth()
        {
            if (this.lsvDevices.Items.Count > 0)
            {
                this.columnMac.Width = -1;
                this.columnName.Width = -1;
                if (this.scaling == 2)
                {
                    this.columnName.Width += 12;
                }
            }
        }

        private void CheckStatus()
        {
            if (this.btInitialized == true)
            {
                EnableControls(false);
                this.statusBar.Text = Resources.StatusBTInitializedOK;
            }
            else
            {
                foreach (Control c in this.Controls)
                {
                    c.Enabled = false;
                }

                this.statusBar.Text = Resources.StatusBTInitializedNG;
            }
        }

        private BluetoothLibNet.BTST_DEVICEINFO[] CreateDeviceInfos()
        {
            return CreateDeviceInfos(2);
        }

        private BluetoothLibNet.BTST_DEVICEINFO[] CreateDeviceInfos(int number)
        {
            BluetoothLibNet.BTST_DEVICEINFO[] result = null;

            if (number > 0)
            {
                result = new BluetoothLibNet.BTST_DEVICEINFO[number];
                for (int i = 0; i < number; i++)
                {
                    result[i] = new BluetoothLibNet.BTST_DEVICEINFO();
                }
            }

            return result;
        }

        private void AddInfo(BluetoothLibNet.BTST_DEVICEINFO info)
        {
            if (info == null)
            {
                return;
            }

            ListViewItem lvi = new ListViewItem(
                new string[] 
                {
                    new string(info.DeviceAddress.TakeWhile(c => c != '\0').ToArray()),
                    new string(info.DeviceName.TakeWhile(c => c != '\0').ToArray())
                });
            lvi.Tag = info;
            switch (info.LocalClass1)
            {
                case BluetoothLibNet.Def.BTCOD_MAJOR_AUDIO:
                    if (info.LocalClass2 == BluetoothLibNet.Def.BTCOD_AUDIO_HEADSET_PROFILE ||
                        info.LocalClass2 == BluetoothLibNet.Def.BTCOD_AUDIO_HANDS_FREE)
                    {
                        lvi.ImageIndex = 3;
                    }
                    else
                    {
                        lvi.ImageIndex = 6;
                    }
                    break;
                case BluetoothLibNet.Def.BTCOD_MAJOR_COMPUTER:
                    if (info.LocalClass2 == BluetoothLibNet.Def.BTCOD_COMPUTER_DESKTOP ||
                        info.LocalClass2 == BluetoothLibNet.Def.BTCOD_COMPUTER_SERVER)
                    {
                        lvi.ImageIndex = 0;
                    }
                    else
                    {
                        lvi.ImageIndex = 4;
                    }
                    break;
                case BluetoothLibNet.Def.BTCOD_MAJOR_IMAGING:
                    if (info.LocalClass2 == BluetoothLibNet.Def.BTCOD_IMAGING_CAMERA ||
                        info.LocalClass2 == BluetoothLibNet.Def.BTCOD_IMAGING_DISPLAY)
                    {
                        lvi.ImageIndex = 9;
                    }
                    else
                    {
                        lvi.ImageIndex = 1;
                    }
                    break;
                case BluetoothLibNet.Def.BTCOD_MAJOR_LAN_ACCESS_POINT:
                    lvi.ImageIndex = 7;
                    break;
                case BluetoothLibNet.Def.BTCOD_MAJOR_PERIPHERAL:
                    lvi.ImageIndex = 8;
                    break;
                case BluetoothLibNet.Def.BTCOD_MAJOR_PHONE:
                    lvi.ImageIndex = 2;
                    break;
                default:
                    lvi.ImageIndex = 5;
                    break;
            }
            this.lsvDevices.Items.Add(lvi);
        }

        private static void PlayBeep()
        {
            try
            {
                if (SystemLibNet.Api.SysPlayBuzzer(
                    SystemLibNet.Def.B_ALARM,
                    SystemLibNet.Def.BUZ_DEFAULT,
                    SystemLibNet.Def.BUZ_DEFAULT) == SystemLibNet.Def.FUNCTION_UNSUPPORT)
                {
                    NativeMethods.MessageBeep(NativeMethods.MB_OK); // IT-3100
                }
            }
            catch (Exception exc)
            {
                Debug.WriteLine(exc.ToString());
            }
        }

        #endregion

    }

    internal class NativeMethods
    {
        internal const int MB_OK = 0x00;
        internal const int MB_ICONHAND = 0x10;
        internal const int MB_ICONQUESTION = 0x20;
        internal const int MB_ICONEXCLAMATION = 0x30;
        internal const int MB_ICONASTERISK = 0x40;

        private NativeMethods() { }

        [DllImport("coredll.dll", SetLastError = false)]
        internal static extern int MessageBeep(int uType);
    }
}